from scipy.stats import multivariate_normal
from torch.utils.data import TensorDataset

from glasflow import RealNVP
import matplotlib.pyplot as plt
import numpy as np
from sklearn.datasets import make_blobs
from sklearn.model_selection import train_test_split
import torch

torch.manual_seed(1451)
#np.random.seed(1451)
plt.style.use('seaborn-colorblind')

def gen_target(density, nside=4, surf_loc=None, loc=None):
#
# function for generating a surface map given density data
#
    
    # if no surface locations rare given then make them
    if surf_loc is None:
        surf_loc = np.zeros((nside**2,3))
        cnt = 0
        for i in np.arange(nside):
            for j in np.arange(nside):
                # place the locations in teh cenres of the grid squares
                surf_loc[cnt,:] = [0.0,i+0.5,j+0.5]
                cnt += 1

    # if no vocel locations are given then make them
    if loc is None:
        loc = np.zeros((nside**3,3))
        cnt = 0
        for i in np.arange(nside):
            for j in np.arange(nside):
                for k in np.arange(nside):
                    # place the locations in the voxel centres
                    loc[cnt,:] = [i+0.5,j+0.5,k+0.5]
                    cnt += 1

    # compute the vertical force at each surface location 
    surf = np.zeros(nside**2)
    for i in range(nside*nside):
        r = loc[:,:] - surf_loc[i,:]   # the vector from the voxel to the surface point
        costheta = r[:,0]/np.sqrt(np.sum(r**2,axis=1))   # the angle between that vector and the z-axis
        surf[i] = np.sum(density*costheta/np.sum(r**2,axis=1))  # the sum of the contributions from all voxels

    return surf

def gen_data(n=100,nside=4,density_sig=1.0,noise_sig=0.05):
#
# function for generating instances of density and surface map pairs
#

    # generate the voxel locations
    loc = np.zeros((nside**3,3))
    cnt = 0
    for i in np.arange(nside):
        for j in np.arange(nside):
            for k in np.arange(nside):
                loc[cnt,:] = [i+0.5,j+0.5,k+0.5]
                cnt += 1

    # generate the surface locations
    surf_loc = np.zeros((nside**2,3))
    cnt = 0
    for i in np.arange(nside):
        for j in np.arange(nside):
            surf_loc[cnt,:] = [0.0,i+0.5,j+0.5]
            cnt += 1

    # generate many instances of densities
    data = []
    targets = []
    norm = ((2.0*np.pi*density_sig*density_sig)**1.5)   # normalisation to make the over-density have unit amplitude
    for j in range(n):
        mu = np.random.uniform(0,nside,3)    # randomly select a mean location for the over-density
        amp = 1.0 #np.random.uniform(0,1,1)       # randomly select the amplitude of the over-density

        # evaluate the density of the 3-D Gaussian at teh voxel centres and add noise
        density = amp*norm*multivariate_normal.pdf(loc,mean=mu,cov=density_sig*density_sig*np.eye(3)) + np.random.normal(0,noise_sig,nside**3)

        # compute the corresponding surface map
        surf = gen_target(density,nside=nside,loc=loc,surf_loc=surf_loc)
        data.append(density)   # append density to full list of events
        targets.append(surf)   # append surface map to full list of events
        if j % 1000 == 0:      # output status to screen
            print('generated {}/{} data'.format(j,n))

    # return flattened density and surface map data
    return np.array(data).reshape(n,-1), np.array(targets).reshape(n,-1)

# generate training and validation data
train_data, train_targets = gen_data(50000)
val_data, val_targets = gen_data(1000)
batch_size = 1000   # define batch size
n_val_plots = 8     # number of validation example plots per epoch

# package the data in a pytorch data loader
x_train_tensor = torch.from_numpy(train_data.astype(np.float32))
y_train_tensor = torch.from_numpy(train_targets.astype(np.float32))
train_dataset = torch.utils.data.TensorDataset(x_train_tensor, y_train_tensor)
train_loader = torch.utils.data.DataLoader(
    train_dataset, batch_size=batch_size, shuffle=True
)
x_val_tensor = torch.from_numpy(val_data.astype(np.float32))
y_val_tensor = torch.from_numpy(val_targets.astype(np.float32))
val_dataset = torch.utils.data.TensorDataset(x_val_tensor, y_val_tensor)
val_loader = torch.utils.data.DataLoader(
    val_dataset, batch_size=batch_size, shuffle=False
)

device = 'cpu'     # choose cpu or gpu (cuda) device

# define flow model
flow = RealNVP(
    n_inputs=64,
    n_transforms=2,
    n_conditional_inputs=16,
    n_neurons=16,
    batch_norm_between_transforms=True
)
flow.to(device)
print(f'Created flow and sent to {device}')

# define optimiser and number of epochs
optimiser = torch.optim.Adam(flow.parameters())
epochs = 400
loss = dict(train=[], val=[])  # initialise loss 

# begin training and loop over epochs
for i in range(epochs+1):
    flow.train()
    train_loss = 0.0
    for batch in train_loader:
        x, y = batch
        x = x.to(device)
        y = y.to(device)
        optimiser.zero_grad()
        _loss = -flow.log_prob(x, conditional=y).mean()
        _loss.backward()
        optimiser.step()
        train_loss += _loss.item()
    loss['train'].append(train_loss / len(train_loader))

    flow.eval()
    val_loss = 0.0
    for batch in val_loader:
        x, y = batch
        x = x.to(device)
        y = y.to(device)
        with torch.no_grad():
            _loss = -flow.log_prob(x, conditional=y).mean().item()
        val_loss += _loss
    loss['val'].append(val_loss / len(val_loader))

    # every n'th epoch print status and make some plots
    if not i % 50:
        print(f"Epoch {i} - train: {loss['train'][-1]:.3f}, val: {loss['val'][-1]:.3f}")
        x_out = []
        with torch.no_grad():
            x, c = next(iter(val_loader))   # get the first validation data
            for k in range(n_val_plots):
                x_out.append(flow.sample(1, c[k].unsqueeze(0)).cpu().numpy())   # generate a density from the flow conditional on the validation surface map

        # make plots
        for k in range(n_val_plots):
            c_out = gen_target(x_out[k],nside=4)   # compute the surface map for the flow output density - should be the same as the conditional surface map
            vmax = np.max([np.max(c_out.flatten()),np.max(c[k].unsqueeze(0).cpu().numpy().flatten())])   # get the maximum surface map value to equally normalise surface plots
            f, ax= plt.subplots(5, 2)
            ax[0,0].imshow(c[k].unsqueeze(0).cpu().numpy().reshape(4,4), cmap='hot',vmin=0.0,vmax=vmax)
            ax[0,0].title.set_text('TRUTH')
            ax[0,0].set_ylabel('surface')
            ax[0,0].set_xticks([])
            ax[0,0].set_yticks([])
            ax[0,1].imshow(c_out.reshape(4,4), cmap='hot',vmin=0.0,vmax=vmax)
            ax[0,1].title.set_text('FLOW')
            ax[0,1].set_ylabel('surface')
            ax[0,1].set_xticks([])
            ax[0,1].set_yticks([])
            for j in range(1,5):
                ax[j,0].imshow(x[k].unsqueeze(0).cpu().numpy().reshape(4,4,4)[j-1,:,:], cmap='binary',vmin=0.0,vmax=1.0)  # gray scale normalised between 0 and 1
                ax[j,1].imshow(x_out[k].reshape(4,4,4)[j-1,:,:], cmap='binary',vmin=0.0,vmax=1.0)
                ax[j,0].set_ylabel('layer {}'.format(j))
                ax[j,1].set_ylabel('layer {}'.format(j))   
                ax[j,0].set_xticks([])
                ax[j,0].set_yticks([])
                ax[j,1].set_xticks([])
                ax[j,1].set_yticks([])
            plt.savefig('/data/www.astro/chrism/henitest/sample_{}_{}.png'.format(i,k))
            plt.close()

flow.eval()
print('Finished training')

plt.figure()
plt.plot(loss['train'][10:], label='Train')
plt.plot(loss['val'][10:], label='Val.')
plt.xlim([10,len(loss['train'])])
plt.xlabel('Epoch')
plt.ylabel('Loss')
plt.savefig('/data/www.astro/chrism/henitest/loss.png')
plt.close()


